use std::collections::HashMap;
use std::fs;
use std::path::Path;
use std::sync::Arc;

use clap::{Parser, Subcommand};
use color_eyre::eyre::WrapErr;
use color_eyre::Result;
use log::LevelFilter;
use walkdir::WalkDir;

use crate::model::coordinator_config::WithFallbackId;

mod coordinator;
mod handler_host;
mod handler_runner;
mod icons;
mod model;

#[derive(Debug, Parser)]
#[command(name = "deckster", about = "Use Loupedeck devices under Linux.")]
struct Cli {
    #[command(subcommand)]
    command: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    Coordinator {
        #[arg(short, long, required = true, alias = "start")]
        config: Box<Path>,
    },
    HandlerHost {
        #[arg(short, long, required = true)]
        config: Box<Path>,
    },
}

#[tokio::main]
pub async fn main() -> Result<()> {
    env_logger::builder().filter_module("deckster", LevelFilter::Info).parse_env("RUST_LOG").init();

    let cli = Cli::parse();

    match cli.command {
        Command::Coordinator { config: config_path } => {
            let deckster_file = read_and_deserialize::<model::coordinator_config::File>(config_path.join("deckster.toml").as_path())?;
            let config_path = config_path.canonicalize()?;

            let key_pages_by_id: HashMap<String, model::key_page::Page> =
                read_and_deserialize_from_directory::<model::key_page::File>(config_path.join("key-pages").as_path())?
                    .into_iter()
                    .map(|p| model::key_page::Page {
                        id: p.inner.id.clone().unwrap_or(p.fallback_id),
                        keys: p.inner.keys.into_iter().map(|(p, k)| (p, Arc::new(k))).collect(),
                        scrolling: p.inner.scrolling,
                    })
                    .map(|p| (p.id.clone(), p))
                    .collect();

            let knob_pages_by_id: HashMap<String, model::knob_page::Page> =
                read_and_deserialize_from_directory::<model::knob_page::File>(config_path.join("knob-pages").as_path())?
                    .into_iter()
                    .map(|p| model::knob_page::Page {
                        id: p.inner.id.clone().unwrap_or(p.fallback_id),
                        knobs: p.inner.knobs.into_iter().map(|(p, k)| (p, Arc::new(k))).collect(),
                    })
                    .map(|p| (p.id.clone(), p))
                    .collect();

            let config = model::coordinator_config::Config {
                active_button_color: deckster_file.active_button_color,
                inactive_button_color: deckster_file.inactive_button_color,
                label_font_family: deckster_file.label_font_family,
                key_pages_by_id,
                knob_pages_by_id,
                buttons: deckster_file.buttons.into_iter().collect(),
                icon_packs: deckster_file.icon_packs,
                initial: deckster_file.initial,
                mqtt: deckster_file.mqtt,
            }
            .validate()?;

            coordinator::start(&config_path, config).await?
        }
        Command::HandlerHost { config: config_path } => {
            let config = read_and_deserialize::<model::handler_host_config::Config>(config_path.join("deckster.toml").as_path())?;
            let config_path = config_path.canonicalize()?;

            handler_host::start(&config_path, config).await?
        }
    };

    Ok(())
}

fn read_and_deserialize<T: serde::de::DeserializeOwned>(path: &Path) -> Result<T> {
    let content = fs::read_to_string(path).wrap_err_with(|| format!("Reading of {} failed.", path.to_string_lossy()))?;

    toml::from_str::<T>(&content).wrap_err_with(|| format!("Parsing of {} failed.", path.to_string_lossy()))
}

fn read_and_deserialize_from_directory<T: serde::de::DeserializeOwned>(base: &Path) -> Result<Vec<WithFallbackId<T>>> {
    let mut result: Vec<WithFallbackId<T>> = Vec::new();

    for entry in WalkDir::new(base).follow_links(true) {
        let entry = entry?;
        if !entry.file_type().is_file() {
            continue;
        }

        let path = entry.path();
        if let Some(fallback_id) = path
            .strip_prefix(base)
            .unwrap()
            .to_str()
            .expect("Paths must be valid UTF-8")
            .strip_suffix(".toml")
        {
            result.push(WithFallbackId {
                fallback_id: fallback_id.to_owned(),
                inner: read_and_deserialize::<T>(path)?,
            });
        }
    }

    Ok(result)
}
